
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/namei.h>
#include <sys/kernel.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/dirent.h>
#include <sys/unistd.h>

#include <vm/vm.h>
#include <vm/vm_zone.h>
#include <vm/vnode_pager.h>

#include "ods2.h"

static u_long ods2hash;

#define DEBUG_HASH 0
#define INOHASH(device, inum)   ((minor(device) + (inum)) & ods2hash)

static struct ods2_node **ods2_hash_table;

#ifndef NULL_SIMPLELOCKS
static struct simplelock ods2_ihash_slock;
#endif  

int
ods2_init(vfsp)
    struct vfsconf *vfsp;
{
    ods2_hash_table = hashinit(desiredvnodes, M_ODS2FSMNT, &ods2hash);
    simple_lock_init(&ods2_ihash_slock);

    return 0;
}

/*
 * Remove the inode from the hash table.
 */
void
ods2_ihashrem(ip)
        register struct ods2_node *ip;
{
        register struct ods2_node *iq;

        simple_lock(&ods2_ihash_slock);
        if ((iq = ip->i_next) != NULL)
                iq->i_prev = ip->i_prev;
        *ip->i_prev = iq;
#ifdef DIAGNOSTIC
        ip->i_next = NULL;
        ip->i_prev = NULL;
#endif
        simple_unlock(&ods2_ihash_slock);
}

struct vnode *
ods2_ihashget(dev, inum)
    dev_t dev;
    ino_t inum;
{
    struct proc *p = curproc;               /* XXX */
    struct ods2_node *ip;
    struct vnode *vp;

loop:

if (DEBUG_HASH)
printf("ods2_ihashget[starting loop]\n");

        simple_lock(&ods2_ihash_slock);

if (DEBUG_HASH)
printf("ods2_ihashget[\tgot ods2_ihash_slock]\n");

        for (ip = ods2_hash_table[INOHASH(dev, inum)]; ip; ip = ip->i_next) {
                if (inum == ip->i_num && dev == ip->i_dev) {

if (DEBUG_HASH)
printf("ods2_ihashget[\tfound %d]\n", inum);

                        vp = ODS2TOV(ip);

if (DEBUG_HASH)
printf("ods2_ihashget[\tlocking vnode interlock]\n");

                        simple_lock(&vp->v_interlock);

if (DEBUG_HASH)
printf("ods2_ihashget[\tunlocking ods2_ihash_slock]\n");

                        simple_unlock(&ods2_ihash_slock);

if (DEBUG_HASH)
printf("ods2_ihashget[\tvget on vp]\n");

                        if (vget(vp, LK_EXCLUSIVE | LK_INTERLOCK, p))
                        {

if (DEBUG_HASH)
printf("ods2_ihashget[\tvget failed, looping again]\n");

                                goto loop;
                        }

if (DEBUG_HASH)
printf("ods2_ihashget[\tvget succeeded, returning]\n");

                        return (vp);
                }
        }

if (DEBUG_HASH)
printf("ods2_ihashget[\tdid not find %d\n", inum);

        simple_unlock(&ods2_ihash_slock);
        return (NULL);
}

void
ods2_ihashins(ip)
        struct ods2_node *ip;
{
        struct proc *p = curproc;               /* XXX */
        struct ods2_node **ipp, *iq;

        simple_lock(&ods2_ihash_slock);
        ipp = &ods2_hash_table[INOHASH(ip->i_dev, ip->i_num)];
        if ((iq = *ipp) != NULL)
                iq->i_prev = &ip->i_next;
        ip->i_next = iq;
        ip->i_prev = ipp;
        *ipp = ip;
        simple_unlock(&ods2_ihash_slock);

        lockmgr(&ip->i_lock, LK_EXCLUSIVE, (struct simplelock *)0, p);
}
